from unittest.mock import patch

import pytest
from infoblox_client import objects

from gso.products.product_types.iptrunk import Iptrunk
from test.conftest import UseJuniperSide
from test.services.conftest import MockedNetboxClient
from test.workflows import (
    assert_complete,
    assert_lso_success,
    extract_state,
    run_workflow,
)


@pytest.fixture()
def trunk_validation_setup(
    request,
    iptrunk_subscription_factory,
    juniper_router_subscription_factory,
    nokia_router_subscription_factory,
    iptrunk_side_subscription_factory,
):
    if request.param == UseJuniperSide.SIDE_A:
        # Nokia -> Juniper
        side_a = iptrunk_side_subscription_factory(iptrunk_side_node=nokia_router_subscription_factory())
        side_b = iptrunk_side_subscription_factory(iptrunk_side_node=juniper_router_subscription_factory())
    elif request.param == UseJuniperSide.SIDE_B:
        # Juniper -> Nokia
        side_a = iptrunk_side_subscription_factory(iptrunk_side_node=juniper_router_subscription_factory())
        side_b = iptrunk_side_subscription_factory(iptrunk_side_node=nokia_router_subscription_factory())
    else:
        # Nokia -> Nokia
        side_a = iptrunk_side_subscription_factory(iptrunk_side_node=nokia_router_subscription_factory())
        side_b = iptrunk_side_subscription_factory(iptrunk_side_node=nokia_router_subscription_factory())

    subscription_id = iptrunk_subscription_factory(iptrunk_sides=[side_a, side_b])
    return [{"subscription_id": subscription_id}]


@pytest.fixture()
def _mocked_netbox_client():
    with (
        patch("gso.services.netbox_client.NetboxClient.get_device_by_name") as mock_get_device_by_name,
        patch("gso.services.netbox_client.NetboxClient.get_available_interfaces") as mock_get_available_interfaces,
        patch("gso.services.netbox_client.NetboxClient.get_available_lags") as mock_get_available_lags,
        patch("gso.services.netbox_client.NetboxClient.create_interface") as mock_create_interface,
        patch("gso.services.netbox_client.NetboxClient.attach_interface_to_lag") as mock_attach_interface_to_lag,
        patch("gso.services.netbox_client.NetboxClient.reserve_interface") as mock_reserve_interface,
        patch("gso.services.netbox_client.NetboxClient.allocate_interface") as mock_allocate_interface,
    ):
        mock_get_device_by_name.return_value = MockedNetboxClient().get_device_by_name()
        mock_get_available_interfaces.return_value = MockedNetboxClient().get_available_interfaces()
        mock_get_available_lags.return_value = MockedNetboxClient().get_available_lags()
        mock_create_interface.return_value = MockedNetboxClient().create_interface()
        mock_attach_interface_to_lag.return_value = MockedNetboxClient().attach_interface_to_lag()
        mock_reserve_interface.return_value = MockedNetboxClient().reserve_interface()
        mock_allocate_interface.return_value = MockedNetboxClient().allocate_interface()

        yield


@pytest.mark.parametrize(
    "trunk_validation_setup", [UseJuniperSide.NONE, UseJuniperSide.SIDE_A, UseJuniperSide.SIDE_B], indirect=True
)
@pytest.mark.workflow()
@pytest.mark.usefixtures("_mocked_netbox_client")
@patch("gso.services.infoblox.find_network_by_cidr")
@patch("gso.services.infoblox.find_v6_host_by_fqdn")
@patch("gso.services.infoblox.find_host_by_fqdn")
@patch("gso.services.lso_client._send_request")
@patch("gso.services.netbox_client.NetboxClient.get_interface_by_name_and_device")
def test_validate_iptrunk_success(
    mock_get_interface_by_name,
    mock_validate_iptrunk,
    mock_find_host_by_fqdn,
    mock_find_v6_host_by_fqdn,
    mock_find_network_by_cidr,
    faker,
    data_config_filename,
    trunk_validation_setup,
):
    #  Mock value setup
    subscription_id = trunk_validation_setup[0]["subscription_id"]
    trunk = Iptrunk.from_subscription(subscription_id).iptrunk
    mock_find_network_by_cidr.side_effects = [
        objects.Network(connector=None, ipv4addrs=[trunk.iptrunk_ipv4_network]),
        objects.NetworkV6(connector=None, ipv6addrs=[trunk.iptrunk_ipv6_network]),
    ]
    mock_find_host_by_fqdn.side_effect = [
        objects.HostRecord(
            connector=None,
            aliases=[trunk.iptrunk_sides[0].iptrunk_side_node.router_fqdn],
            comment=subscription_id,
            ipv4addrs=[
                objects.IPv4(
                    ipv4addr=str(trunk.iptrunk_ipv4_network[0]),
                    configure_for_dhcp=False,
                    mac="00:00:00:00:00:00",
                    ip=str(trunk.iptrunk_ipv4_network[0]),
                    host=f"{trunk.iptrunk_sides[0].iptrunk_side_ae_iface}.{trunk.iptrunk_sides[0].iptrunk_side_node.router_fqdn}",
                ),
            ],
            name=trunk.iptrunk_sides[0].iptrunk_side_node.router_fqdn,
        ),
        objects.HostRecord(
            connector=None,
            aliases=[trunk.iptrunk_sides[1].iptrunk_side_node.router_fqdn],
            comment=subscription_id,
            ipv4addrs=[
                objects.IPv4(
                    ipv4addr=str(trunk.iptrunk_ipv4_network[1]),
                    configure_for_dhcp=False,
                    mac="00:00:00:00:00:00",
                    ip=str(trunk.iptrunk_ipv4_network[1]),
                    host=f"{trunk.iptrunk_sides[1].iptrunk_side_ae_iface}.{trunk.iptrunk_sides[1].iptrunk_side_node.router_fqdn}",
                ),
            ],
            name=trunk.iptrunk_sides[1].iptrunk_side_node.router_fqdn,
        ),
    ]

    mock_find_v6_host_by_fqdn.side_effect = [
        objects.HostRecordV6(
            connector=None,
            aliases=[trunk.iptrunk_sides[0].iptrunk_side_node.router_fqdn],
            comment=subscription_id,
            ipv6addrs=[
                objects.IPv6(
                    ipv6addr=str(trunk.iptrunk_ipv6_network[1]),
                    configure_for_dhcp=False,
                    mac="00:00:00:00:00:00",
                    ip=str(trunk.iptrunk_ipv6_network[1]),
                    host=f"{trunk.iptrunk_sides[0].iptrunk_side_ae_iface}.{trunk.iptrunk_sides[0].iptrunk_side_node.router_fqdn}",
                ),
            ],
            name=trunk.iptrunk_sides[0].iptrunk_side_node.router_fqdn,
        ),
        objects.HostRecordV6(
            connector=None,
            aliases=[trunk.iptrunk_sides[1].iptrunk_side_node.router_fqdn],
            comment=subscription_id,
            ipv6addrs=[
                objects.IPv6(
                    ipv6addr=str(trunk.iptrunk_ipv6_network[2]),
                    configure_for_dhcp=False,
                    mac="00:00:00:00:00:00",
                    ip=str(trunk.iptrunk_ipv6_network[2]),
                    host=f"{trunk.iptrunk_sides[1].iptrunk_side_ae_iface}.{trunk.iptrunk_sides[1].iptrunk_side_node.router_fqdn}",
                ),
            ],
            name=trunk.iptrunk_sides[1].iptrunk_side_node.router_fqdn,
        ),
    ]

    mock_get_interface_by_name.side_effect = [
        MockedNetboxClient.BaseMockObject(
            name="Boeiee",
            module=MockedNetboxClient.BaseMockObject(display="display1"),
            description=subscription_id,
            enabled=True,
        ),
        MockedNetboxClient.BaseMockObject(
            name="2222222",
            module=MockedNetboxClient.BaseMockObject(display="display2"),
            description=subscription_id,
            enabled=True,
        ),
        MockedNetboxClient.BaseMockObject(
            name="3333",
            module=MockedNetboxClient.BaseMockObject(display="display 333"),
            description=subscription_id,
            enabled=True,
        ),
        MockedNetboxClient.BaseMockObject(
            name="44",
            module=MockedNetboxClient.BaseMockObject(display="display 444"),
            description=subscription_id,
            enabled=True,
        ),
        MockedNetboxClient.BaseMockObject(
            name="55",
            module=MockedNetboxClient.BaseMockObject(display="display25555"),
            description=subscription_id,
            enabled=True,
        ),
        MockedNetboxClient.BaseMockObject(
            name="666",
            module=MockedNetboxClient.BaseMockObject(display="display666"),
            description=subscription_id,
            enabled=True,
        ),
    ]

    #  Run workflow
    initial_router_data = [{"subscription_id": subscription_id}]
    result, process_stat, step_log = run_workflow("validate_iptrunk", initial_router_data)

    state = extract_state(result)

    for _ in range(3):
        result, step_log = assert_lso_success(result, process_stat, step_log)

    assert_complete(result)

    subscription_id = state["subscription_id"]
    subscription = Iptrunk.from_subscription(subscription_id)

    assert subscription.status == "active"
    assert mock_validate_iptrunk.call_count == 3
    assert mock_find_host_by_fqdn.call_count == 2
    assert mock_find_v6_host_by_fqdn.call_count == 2
    assert mock_find_network_by_cidr.call_count == 2


@pytest.mark.workflow()
@pytest.mark.usefixtures("_mocked_netbox_client")
@patch("gso.services.infoblox.find_network_by_cidr")
@patch("gso.services.infoblox.find_v6_host_by_fqdn")
@patch("gso.services.infoblox.find_host_by_fqdn")
@patch("gso.services.lso_client._send_request")
@patch("gso.services.netbox_client.NetboxClient.get_interface_by_name_and_device")
def test_validate_iptrunk_skip_legacy_trunks(
    mock_get_interface_by_name,
    mock_validate_iptrunk,
    mock_find_host_by_fqdn,
    mock_find_v6_host_by_fqdn,
    mock_find_network_by_cidr,
    faker,
    data_config_filename,
    iptrunk_subscription_factory,
    iptrunk_side_subscription_factory,
    juniper_router_subscription_factory,
):
    #  Mock value setup
    side_a = iptrunk_side_subscription_factory(iptrunk_side_node=juniper_router_subscription_factory())
    side_b = iptrunk_side_subscription_factory(iptrunk_side_node=juniper_router_subscription_factory())
    subscription_id = iptrunk_subscription_factory(iptrunk_sides=[side_a, side_b])

    #  Run workflow
    initial_router_data = [{"subscription_id": subscription_id}]
    result, _, _ = run_workflow("validate_iptrunk", initial_router_data)
    state = extract_state(result)
    assert_complete(result)

    subscription_id = state["subscription_id"]
    subscription = Iptrunk.from_subscription(subscription_id)

    assert subscription.status == "active"
    assert mock_get_interface_by_name.call_count == 0
    assert mock_validate_iptrunk.call_count == 0
    assert mock_find_host_by_fqdn.call_count == 0
    assert mock_find_v6_host_by_fqdn.call_count == 0
    assert mock_find_network_by_cidr.call_count == 0
